windowcontrols.overlaid > button > image,
windowcontrols.overlaid > button,
windowcontrols.overlaid > button:hover,
.overlaid button,
button.overlaid {
  box-shadow: initial;
}

windowcontrols.overlaid > button:hover > image,
.overlaid button:hover,
button.overlaid:hover {
  box-shadow: 0 0 0 1px white;
}

.shade {
  background-image: linear-gradient(
    alpha(black, 0.5) 0%,
    alpha(black, 0.4) 10%,
    alpha(black, 0.35) 20%,
    alpha(black, 0.3) 30%,
    alpha(black, 0.3) 65%,
    alpha(black, 0.35) 75%,
    alpha(black, 0.45) 85%,
    alpha(black, 0.6) 100%
  );
}
